#!/bin/ksh
#
# hwrf_rtofs_init.sh prepares the ocean files for the hurricane model
# the 3 main functions is does is create
#   - boundary conditions - these may be from the rtofs full domain run, or
#                           a mixture of the rtofs and the previous cycle
#                           of the hurricane run.
#   - initial conditions - restart file from archive file
#   - forcing files - from gfs flux (melded?) 
#

set -x

echo "*** Started script $0"

##### SET DATE AND OTHER ENVIRONMENTAL VARIABLE FROM THE SUPER SCRIPT #######
##### GET THE INPUT VARIABLES

if [ $# -eq 6 ]
then
   export DOMAIN=${1}
   export YEAR=${2}
   export MONTH=${3}
   export DATE=${4}
   export CYCLE=${5}
   export STORMID=${6}

   echo 'DOMAIN  = ' ${DOMAIN}
   echo 'YEAR    = ' ${YEAR}
   echo 'MONTH   = ' ${MONTH}
   echo 'DATE    = ' ${DATE}
   echo 'CYCLE   = ' ${CYCLE}
   echo 'STORMID = ' ${STORMID}
else
   echo 'Usage: sh hwrf_rtofs_init.sh DOMAIN YEAR MONTH DATE CYCLE STORMID'
   echo 'NEED SIX INPUT VARIABLES'
   echo 'SCRIPT WILL EXIT'
   exit 1
fi

##############################################################
export pdy=$YEAR$MONTH$DATE
export cyc=$CYCLE
export LCASE_NAME=`echo  $DOMAIN | tr '[A-Z]' '[a-z]' `
name=$LCASE_NAME

echo 'pdy      = ' ${pdy}
echo 'cyc      = ' ${cyc}
echo 'name     = ' ${name}
export prevcycle=`${utilexec}/ndate -6 $YYYY$MM$DD$CYC`
cycm1=`echo $prevcycle | cut -c9-10`

# find forecast_hours
# --hsk
export spinlength=${spinlength:-24}
forecast_length=${forecast_length:-126}
let fsecs=$forecast_length*3600
secs=${C_NSECS:-$fsecs}
let forecast_hours=$secs/3600

enddate=`${utilexec}/ndate $forecast_hours $YEAR$MONTH$DATE$CYC`
ocean_status=0
oceanstatusfile=${COMOUT}/ocean_status.${DOMAIN}${STORMID}.$YEAR$MONTH$DATE$CYCLE

# determine what needs to be done for this run
# ex. 2005082912/12l/KATRINA.2005082912/OCEAN

#----             Check Ocean is Staged or not --------------------
#------------------------------------------------------------------
OCEAN_staged=${OCEAN_staged:-NO}
if [ "$OCEAN_staged" == YES ]
then
  OCEAN_full=$STAGE_DIR/$STORMID/$YEAR$MONTH$DATE$CYCLE/OCEAN
  if [[ -d $OCEAN_full &&
        -d $OCEAN_full/forcing &&
        -d $OCEAN_full/nest &&
        -d $OCEAN_full/restarts ]]
  then
     mkdir -p $CSTREAM/OCEAN
     ln -sf $OCEAN_full/forcing $CSTREAM/OCEAN/.
     ln -sf $OCEAN_full/nest $CSTREAM/OCEAN/.
     ln -sf $OCEAN_full/restarts  $CSTREAM/OCEAN/.
     ln -sf $OCEAN_full/${Application} $CSTREAM/OCEAN/.
     cp $OCEAN_full/domain_name $CSTREAM/OCEAN/.

     oceanstatusfile=${COMOUT}/ocean_status.${DOMAIN}${STORMID}.$YEAR$MONTH$DATE$CYCLE
     echo 'set flag for a coupled model run'
     echo "export RUN_COUPLED=YES"      >> ${oceanstatusfile}
     echo "export OCEAN_STATUS=1"       >> ${oceanstatusfile}
     # cp $OCEAN_full/zeroth_file/* $COMOUT

     echo $OCEAN_full
     echo is present and assumed to be populated
     echo rtofs_init is returning with return code 0
  
     exit 0
  fi
fi

###############################################################
# OCEAN_staged = NO:
#     Create BC, IC and Forcings
#--------------------------------------------------------------
ocean_status=0
boundary_conditions_from_rtofs=T
     
force_coldstart=${force_coldstart-:0}
if [[ $same_domain = 1 && $force_coldstart = 0 ]]
then
   spin=T
   create_subdomain=F
else
   spin=F
   create_subdomain=T
fi
  
mkdir -p $CSTREAM/OCEAN
echo $PDY: Setting-up BC, IC and Forcings

#################################################################################
# 1. BOUNDARY CONDITIONS

# 1.1a create subdomain from rtofs

boundary_conditions_from_rtofs=${boundary_conditions_from_rtofs:-T}
if [[ $create_subdomain = T || $boundary_conditions_from_rtofs = T ]]
then

# determine set of archives that subset_bdry needs to create the BC files. 
  $USHhwrf/hwrf_rtofs_subset_bdry_init.sh $spinlength $forecast_length
  echo $Application > $CSTREAM/OCEAN/domain_name

else

# 1.1b boundary conditions for this cycle is combination of hwrf and rtofs
  # hwrf - copy as much as we can from previous cycle of hwrf
  echo SAME DOMAIN - bc from hwrf + rtofs
  
  mkdir -p $CSTREAM/OCEAN/$Application
  echo $Application > $CSTREAM/OCEAN/domain_name

  typeset -Z4 lastpy
  lastpy=0000
  
  typeset -Z3 lastpd
  lastpd=000

  typeset -Z2 lastph
  lastph=00
 
  icd=0

  while [ $icd -le $forecast_hours ]
  do
    # icd is hour for previous cycle, im6 is hour for this cycle
    let im6=icd-6
    prev=`${utilexec}/ndate $im6 $YEAR$MONTH$DATE$CYCLE`
    pymd=`echo $prev | cut -c1-8`
    py=`echo $prev | cut -c1-4`
    ph=`echo $prev | cut -c9-10`
    pyddd=`$utilscript/date2jday.sh $pymd`
    pd=`echo $pyddd | cut -c5-7`
  
# copy the input files....
    if [ -a $HISTDATA/hwrf_$RUNmodIDout.t${cycm1}z.${py}_${pd}_${ph}.a ]
    then
       cp -p $HISTDATA/hwrf_$RUNmodIDout.t${cycm1}z.${py}_${pd}_${ph}.a $CSTREAM/OCEAN/$Application/hwrf_$RUNmodIDout.t00z.${py}_${pd}_${ph}.a
       cp -p $HISTDATA/hwrf_$RUNmodIDout.t${cycm1}z.${py}_${pd}_${ph}.b $CSTREAM/OCEAN/$Application/hwrf_$RUNmodIDout.t00z.${py}_${pd}_${ph}.b
       lastpy=${py}
       lastpd=${pd}
       lastph=${ph}
    else
       echo hwrf_$RUNmodIDout.t${cycm1}z.${py}_${pd}_${ph}.[ab] does not exist, breaking loop
       break
    fi 

    let icd=icd+1
  done
  # rtofs - if need more, then get the rest from rtofs
  echo copied about $im6 files - need files after $lastpy $lastpd $lastph
  nymd=`${utilscript}/date2jday.sh $lastpy$lastpd`
  nymdh=`${utilexec}/ndate 1 $nymd$lastph`

  if [ $nymdh -lt $enddate ]
  then 
    echo $PDY: subsetting for rtofs to fill in missing data
    diffhours=`${utilexec}/nhour ${enddate} ${nymdh}`
    $USHhwrf/hwrf_rtofs_subset_bdry_init.sh 0 $diffhours
  fi
fi

# 1.2 populate the boundary conditions in nest directory
# link in each nested archive file as correct name
# was hwrf_ofs_hwrf04.t00z.2010_263_18.? to archv_2010_263_18.?
# now hwrf_rtofs_hat11.t00z.f01.?        to archv_f01.?
# dan archv_f01.? does not work. need to get yyyy_ddd_hh 

mkdir -p ${CSTREAM}/OCEAN/nest
cd ${CSTREAM}/OCEAN/$Application
for bfile in `ls -C1 hwrf_${RUNmodIDout}.*.b`
do
    filroot=`echo $bfile | cut -d. -f1-2`
    afile=$filroot.a

    modelday=`sed -n 11,11p $bfile | awk '{print $4}'`
    mydh=`sh $utilscript/date_hycom2normal.sh $modelday`
 
    myd=`echo $mydh | cut -c1-8`
    yd=`$utilscript/date2jday.sh $myd`
    myyyy=`echo $yd | cut -c1-4`
    mddd=`echo $yd | cut -c5-7`
    mhh=`echo $mydh | cut -c9-10`
    ln -s ../$Application/$afile  ${CSTREAM}/OCEAN/nest/archv.${myyyy}_${mddd}_${mhh}.a
    ln -s ../$Application/$bfile  ${CSTREAM}/OCEAN/nest/archv.${myyyy}_${mddd}_${mhh}.b
done

# check if minimum number of BC files created
nknt=`echo ls ${CSTREAM}/OCEAN/${Application}/*rtofs*.a | wc | cut -c10-15`
if [[ $nknt -ge 48 ]];
then
     let ocean_status=ocean_status + 1
else
     echo "Setting BC Failed "
     msg="ABNORMAL EXIT RTOFS INIT: Not enough BC files"
     postmsg "$jlogfile" "$msg"
     err=99;
     export err;err_chk
fi

#################################################################################
# 2. INITIAL CONDITIONS

###########
# if $SPINUP_DATA/*restart*n00 is same as YEARMONTHDATECYCLE
#   then run restart2restart (subregioning) to this domain for the IC
#   else
#     if $SPINUP_DATA/*restart*n00 is same as YEARMONTHDATECYCLE-6
#     then run restart2restart (subregioning) to this domain and use for spin
#     else run archv2restart on archive from $SPINUP_DATA from YEARMONTHDATECYCLE-6 and use for spin
#
# new logic:
# run restart2restart on $SPINUP_DATA/*restart*n00
# if this restart file is not same time as YEARMONTHDATECYCLE-6
#   then run archv2restart on archive from $SPINUP_DATA for YEARMONTHDATECYCLE-6 
# use created restart to spin for 6 hours
###########

#
mkdir -p $CSTREAM/OCEAN/restarts
#
echo SPINNING an analysis
$USHhwrf/hwrf_rtofs_spin.sh
spinrc=$?

ocean_status=0

# if spin fails, then get restart file from rtofs
#dan: what if spin fails???!!! here we get 6hour restart from coupled forecast (not available in global)
if [ $spinrc != 0 ]
then
    # postmsg
    echo spin failed, what to do now?
    cp -p $HISTDATA/hwrf_$RUNmodIDout.t${cycm1}z.nest_restartR.a $CSTREAM/OCEAN/restarts/restart_pre.a
    cp -p $HISTDATA/hwrf_$RUNmodIDout.t${cycm1}z.nest_restartR.b $CSTREAM/OCEAN/restarts/restart_pre.b
else
   let ocean_status=ocean_status+1
# copy restarts to comout for next cycle spin
  cp $CSTREAM/OCEAN/restarts/restart_out.a $COMOUT/hwrf_${RUNmodIDout}.t${cyc}z.spin_restart.a
  cp $CSTREAM/OCEAN/restarts/restart_out.b $COMOUT/hwrf_${RUNmodIDout}.t${cyc}z.spin_restart.b
fi

#################################################################################
# fill in flags - RUN_COUPLED=YES/NO and OCEAN_STATUS=1/0
#----------------------------------------------------------
#if [ ${ocean_status} -ne 0 ]
#then

  # ocean initial and boundary conditions were formed: 
  # prepare forcing files to model to will run coupled 

  #################################################################################
  # 3. FORCING FILES 

  # create forcing files for the coupled forecast step (6 days?)
  # adjust_river - call ofs_create_river.sh to update climatology river with USGS data.
  # adjust_temp - call ofs_correct_forcing.sh
  # adjust_wind - run hsk wind algorithm, =1 create them, =2 use them

#  adjust_wind=0

#  interval=${interval:-3}
#  adjust_river=${adjust_river:-0}
#  adjust_temp=${adjust_temp:-0}
#  adjust_wind=${adjust_wind:-0}
#  export mode=${mode:-fcst}

#  mkdir ${DATA}/forcing
#  cd ${DATA}/forcing
#  $utilscript/setup.sh

#  $USHhwrf/hwrf_rtofs_get_forcing.sh $PDY$CYC $enddate $interval $adjust_river $adjust_temp $adjust_wind
#  scriptrc=$?

#  if [ $scriptrc -eq 0 ]
#  then
#     mkdir -p ${CSTREAM}/OCEAN/forcing
#     cp -p forcing.*.? ${CSTREAM}/OCEAN/forcing
  
#     let ocean_status=ocean_status+1
#  fi
#fi

###################################################################
# Diagnostics
#------------
if [ $ocean_status -ne 0 ]
then
   echo '--- HWRF_RTOFS_INIT.SH --- '
   echo ' IC, BC, and FORCING are ready and '
   echo ' Set flag for a coupled model run'
      
   ocean_status=1   
   
   echo "export RUN_COUPLED=YES"               >> ${oceanstatusfile}
   echo "export OCEAN_STATUS=1"  >> ${oceanstatusfile}

   echo "init1.complete" > $CSTREAM/init/init1.complete

else
   echo '--- HWRF_RTOFS_INIT.SH --- '
   echo ' A coupled model run is NOT ready '
  
  # ocean initial and boundary conditions were not formed: 
  # model will run uncoupled
  echo 'set flag for an atmosphere model only run'
  echo "export RUN_COUPLED=NO"                > ${oceanstatusfile}
  echo "export OCEAN_STATUS=0"  >> ${oceanstatusfile}
fi

echo "*** Finished script $0"

